/*
 * $Id: kedge.cc,v 1.2 2002/04/07 16:50:23 jpr5 Exp $
 *
 * kedge 1.1
 * Copyright (c) 2002 Jordan Ritter
 * Released into the Public Domain.
 */

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>

#include <netwm_def.h>
#include <netwm.h>

#include <X11/Xlib.h>


static int (*old_XErrorHandler)(Display*, XErrorEvent *);
static unsigned int x_edge, y_edge;
static int max_desktop, dest_desktop;
static Display *display;
static NETRootInfo *r;
static Window root;
static int factor;


// dummy error handler to deal with the X11 server giving us bogus
// CreateWindow events (on destroy), and other bogosities.
int my_XErrorHandler(Display *d, XErrorEvent *e) {
  if (e->error_code == BadWindow)
    return 0;
  else return old_XErrorHandler(d, e);
}


// not used 
void FlushMouseEvents(Display *d) {
  XEvent e;
  while (XCheckMaskEvent(d, MotionNotify, &e) == True);
}

// KDE_VERSION exists because versions 2.x <-> 3.x differ in how the
// desktops are ordered.
void DetectEdge(unsigned x, unsigned y) {
  if ( x == 0 ) {

    r->activate();
#if KDE_VERSION == 2
    dest_desktop = r->currentDesktop() - max_desktop/factor;
#else
    dest_desktop = r->currentDesktop() - 1;
#endif
    
    if (dest_desktop < 1) 
	return;

    r->setCurrentDesktop(dest_desktop);
    XWarpPointer(display, None, root, 0, 0, 0, 0, x_edge-10, y);

  } else if ( x == x_edge ) {

    r->activate();
#if KDE_VERSION == 2
    dest_desktop = r->currentDesktop() + max_desktop/factor;
#else
    dest_desktop = r->currentDesktop() + 1;
#endif    

    if (dest_desktop > max_desktop) 
      return;

    r->setCurrentDesktop(dest_desktop);
    XWarpPointer(display, None, root, 0, 0, 0, 0, 10, y);

  } else if (factor > 1) {

    if ( y == 0 ) {

      r->activate();
#if KDE_VERSION == 2
      dest_desktop = r->currentDesktop() - 1;
#else
      dest_desktop = r->currentDesktop() - max_desktop/factor;
#endif      

      if (dest_desktop < 1) 
	return;

      r->setCurrentDesktop(dest_desktop);
      XWarpPointer(display, None, root, 0, 0, 0, 0, x, y_edge-10);

    } else if ( y == y_edge ) {

      r->activate();
#if KDE_VERSION == 2
      dest_desktop = r->currentDesktop() + 1;
#else
      dest_desktop = r->currentDesktop() + max_desktop/factor;
#endif

      if (dest_desktop > max_desktop) 
	return;

      r->setCurrentDesktop(dest_desktop);
      XWarpPointer(display, None, root, 0, 0, 0, 0, x, 10);

    }

  }    
}


void setupKDE(Display *d) {
  r = new NETRootInfo(d, NET::NumberOfDesktops|
		         NET::DesktopGeometry|
		         NET::CurrentDesktop|
		         NET::VirtualRoots);
  r->activate();
  max_desktop = r->numberOfDesktops();

  factor = KDE_PAGER_SIZE;
}


void spy(Window r) {
  Window p, *children;
  unsigned int nchildren;

  XWindowAttributes a = {0,};
  XGetWindowAttributes(display, r, &a);
  XSelectInput(display, r, a.your_event_mask|PointerMotionMask);

  if (!XQueryTree(display, r, &r, &p, &children, &nchildren)) {
    fprintf(stderr, "warning: couldn't query window tree\n");
    return;
  } 
    
  if (!nchildren) return;

  for (int i = 0; i < nchildren; i++) 
    spy(children[i]);
  
  XFree(children);
}


int main(void) {
  if (fork())
    return 1;

  old_XErrorHandler = XSetErrorHandler(my_XErrorHandler);

  if (!(display = XOpenDisplay(":0.0"))) {
    fprintf(stderr, "fatal: couldn't open display\n");
    return 1;
  }

  root = DefaultRootWindow(display);

  setupKDE(display);
  
  // find out how big the screen is
  XWindowAttributes a = {0,};
  XGetWindowAttributes(display, root, &a);

  x_edge = a.width-1; 
  y_edge = a.height-1;

  // notify us of changes to any children 
  XSelectInput(display, root, a.your_event_mask|SubstructureNotifyMask);

  // spy on all child windows
  spy(root);

  while (1) {
    XEvent e;
    XNextEvent(display, &e);

    switch( e.type ) {
      case CreateNotify: { // snag events from new windows
	XCreateWindowEvent *ce = (XCreateWindowEvent*)&e;
	XWindowAttributes a = {0,};
	XGetWindowAttributes(display, ce->window, &a); 
	XSelectInput(display, ce->window, a.your_event_mask|PointerMotionMask);
      } break;
      case MotionNotify: {
	XMotionEvent *me = (XMotionEvent*)&e;
	DetectEdge(me->x_root, me->y_root);
      } break;
    }
  }
  
  return 0;
}

